/*
  C64 packer.

  block header's bits 1-0 indicate block tpye
    x 0 unpacked run
    0 1 RLE, bits 7-2 are length, repeat next char length times.
    1 1 equal strings, bits 7-2 are length, next word is backindex.

  Wotan mit uns.
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>


// Defining this will result in some ugly statistics displayed.
//#define MAKE_LOG


#define MAX_BUFLEN      0xfff0L
#define MAX_UNPACKEDLEN 127
#define MAX_RLELEN      63
#define MAX_BACKINDEX   65535L
#define MAX_STRINGLEN   63


#ifndef FALSE
#define FALSE               0
#endif

#ifndef TRUE
#define TRUE                1
#endif


typedef int BOOL;
typedef unsigned char BYTE;             // exactly 8 bit unsigned
typedef unsigned short WORD;            // at least 16 bit unsigned


typedef struct {
  WORD next;                            // index of next array element.
  WORD index;                           // index in inbuf[];
} BYTEINDEX;


class Packer {
  BYTE *inbuf;
  BYTE *outbuf;
  long inbufpos;
  long inbuflen;
  long outbuflen;

  long stringcounter[MAX_STRINGLEN+1];
  long lastindexed;
  WORD byteindices[256];
  BYTEINDEX *byteindexpool;
  WORD byteindexpoolpos;

  void updateByteIndices(long index);

  long getMaxRLELength(long start);
  long getMaxStringLength(long start, long& backindex);

  void writeByte(BYTE b);
  void writeUnpackedBlock(long len);
  void writeRLEBlock(long len);
  void writeStringBlock(long stringlen, long backindex);

public:
  Packer(void);
  ~Packer(void);

  void load(FILE *f, BOOL skipheader=0);
  void pack(void);
  void depack(void);
  void save(FILE *f);
};


Packer::Packer(void)
{
  inbuf=new BYTE[MAX_BUFLEN];
  outbuf=new BYTE[MAX_BUFLEN];
}


Packer::~Packer(void)
{
  delete inbuf;
  delete outbuf;
}


void Packer::updateByteIndices(long index)
{
  BYTE thisbyte;

  while (lastindexed<index) {
    lastindexed++;
    thisbyte=inbuf[lastindexed];
// Register this byte into the byte pool.
    byteindexpool[byteindexpoolpos].next=byteindices[thisbyte];
    byteindexpool[byteindexpoolpos].index=lastindexed;
    byteindices[thisbyte]=byteindexpoolpos;
    byteindexpoolpos++;
  }
}


long Packer::getMaxRLELength(long start)
{
  long len, end;
  BYTE c;

  end=inbuflen-start;
  if (end>MAX_RLELEN) end=MAX_RLELEN;
  c=inbuf[start];
  for (len=1; len<end; len++) {
    if (c!=inbuf[start+len]) break;
  }
  return len;
}


long Packer::getMaxStringLength(long start, long& backindex)
{
  long l;
  long maxlen;
  long bestlen, bestindex;
  WORD byteindex;
  long oldindex;
  BYTE *curr, *old;

  bestlen=0;
  bestindex=0;
  byteindex=byteindices[inbuf[start]];
  while (byteindex!=0xffff) {
    oldindex=byteindexpool[byteindex].index;
    curr=inbuf+start;
    old=inbuf+oldindex;
    if (curr-old<=MAX_BACKINDEX) {
      maxlen=start-oldindex;
      if (maxlen>MAX_STRINGLEN) maxlen=MAX_STRINGLEN;
      if (maxlen>inbuflen-start) maxlen=inbuflen-start;   // Don't go past EOF.
      for (l=0; l<maxlen; l++, curr++, old++) {
        if ((*curr)!=(*old)) break;
      }
      if (bestlen<l) {
        bestlen=l;
        bestindex=oldindex;
      }
    }
    byteindex=byteindexpool[byteindex].next;
  }
  backindex=start-bestindex;
  return bestlen;
}


void Packer::writeByte(BYTE b)
{
  outbuf[outbuflen]=b;
  outbuflen++;
}


void Packer::writeUnpackedBlock(long len)
{
  long i;

  writeByte(len<<1);
  for (i=0; i<len; i++) {
    writeByte(inbuf[inbufpos]);
    inbufpos++;
  }
}


void Packer::writeRLEBlock(long rlelen)
{
  writeByte((rlelen<<2)+1);
  writeByte(inbuf[inbufpos]);
  inbufpos+=rlelen;
}


void Packer::writeStringBlock(long stringlen, long backindex)
{
  writeByte((stringlen<<2)+3);
  writeByte(backindex&0xff);
  writeByte((backindex>>8)&0xff);
  inbufpos+=stringlen;
}


void Packer::pack(void)
{
  int b;
  long s, end;
  long rlelen, stringlen, backindex;

  inbufpos=0;
  outbuflen=0;

// Init byte index lists.
  for (b=0; b<256; b++) byteindices[b]=0xffff;
  lastindexed=-1;
  byteindexpool=new BYTEINDEX[inbuflen];
  byteindexpoolpos=0;

  for (s=0; s<MAX_STRINGLEN+1; s++) {
    stringcounter[s]=0;
  }

  for (;;) {
    if (inbuflen>0) fprintf(stderr, "%6.2f%%\r", (100.0f*inbufpos)/inbuflen);
    end=inbuflen-inbufpos;
    if (end==0) break;
    if (end<0) {
      fprintf(stderr, "Bug in packer. Burn in hell.\n");
      exit(EXIT_FAILURE);
    }
    if (end>MAX_UNPACKEDLEN) end=MAX_UNPACKEDLEN;
// Try to find a run of 3 equal bytes.
    rlelen=0;
    stringlen=0;
    for (s=0; s<end; s++) {
      updateByteIndices(inbufpos+s);
      rlelen=getMaxRLELength(inbufpos+s);
      if (rlelen==MAX_RLELEN) break;
      stringlen=getMaxStringLength(inbufpos+s, backindex);
      if (rlelen>=3) break;
      if (stringlen>=4) break;
    }
    if (s) writeUnpackedBlock(s);
    if (rlelen>=stringlen) {
      if (rlelen>=3) writeRLEBlock(rlelen);
    }
    else {
      if (stringlen>=4) writeStringBlock(stringlen, backindex);
    }
    stringcounter[stringlen]++;
#ifdef MAKE_LOG
    fprintf(stdout, "%6ld %4ld\n", backindex, stringlen);
#endif
  }
  writeUnpackedBlock(0);
  if (inbuflen>0) fprintf(stderr, "%6.2f%%\n", (100.0f*outbuflen)/inbuflen);

#ifdef MAKE_LOG
  for (s=0; s<MAX_STRINGLEN+1; s++) {
    printf("%03ld %04ld\n", s, stringcounter[s]);
  }
#endif

// Free byte index lists.
  delete byteindexpool;
}


void Packer::depack(void)
{
  long i, l, bi;
  BYTE c;

  inbufpos=0;
  outbuflen=0;

  for (;;) {
    c=inbuf[inbufpos];
    if (c==0) break;
    inbufpos++;
// Uncompressed block.
    if ((c&1)==0) {
      l=c>>1;
      for (i=0; i<l; i++) {
        outbuf[outbuflen]=inbuf[inbufpos];
        inbufpos++;
        outbuflen++;
      }
    }
    else {
// RLE block.
      if ((c&2)==0) {
        l=c>>2;
        c=inbuf[inbufpos];
        inbufpos++;
        for (i=0; i<l; i++) {
          outbuf[outbuflen]=c;
          outbuflen++;
        }
      }
// String block.
      else {
        l=c>>2;
        bi=inbuf[inbufpos];
        inbufpos++;
        bi|=inbuf[inbufpos]<<8;
        inbufpos++;
        for (i=0; i<l; i++) {
          outbuf[outbuflen]=outbuf[outbuflen-bi];
          outbuflen++;
        }
      }
    }
  }
}


void Packer::load(FILE *f, BOOL skipheader)
{
  if (skipheader) fseek(f, 2, SEEK_CUR);
  inbuflen=fread(inbuf, 1, MAX_BUFLEN, f);
}


void Packer::save(FILE *f)
{
  fwrite(outbuf, 1, outbuflen, f);
}


int main(int argc, char *argv[])
{
  FILE *in;
  FILE *out;
  Packer *p;

  if (argc<2) {
    puts("argv[1]: command");
    puts("         e => pack");
    puts("         es => skip load address (2 bytes in input) and pack");
    puts("         d => depack");
    puts("argv[2]: input file name");
    puts("argv[3]: output file name");
  }
  else {
    p=new Packer;
    if (argc>2) in=fopen(argv[2], "rb");
    else in=stdin;
    if (in==NULL) {
      fprintf(stderr, "Unable to open input.");
      return EXIT_FAILURE;
    }
    if (argv[1][1]=='s') p->load(in, TRUE);
    else p->load(in);
    if (argc>2) fclose(in);
    if (argv[1][0]=='e') p->pack();
    if (argv[1][0]=='d') p->depack();
    if (argc>3) out=fopen(argv[3], "wb");
    else out=stdout;
    if (out==NULL) {
      fprintf(stderr, "Unable to open output.");
      return EXIT_FAILURE;
    }
    p->save(out);
    if (argc>3) fclose(out);
    delete p;
  }

  return EXIT_SUCCESS;
}
