/*
  TGA to C64 font converter.

  Characters are read from 8x8 pixel boxes from the TGA.
  Char 0 starts at x=0,y=0,
  Char 1 starts at x=8,y=0,
  ...
  Char 31 starts at x=248,y=0,
  Char 32 starts at x=0,y=8,
  ...
  Char 63 starts at x=248,y=8.
  ...

  which totals a 256x16 pixel TGA.
  Maybe it works with any other dimensions, never tried.
  BTW, maybe it can also convert to C64 bitmap
  when the input is 320 pixels wide.

  argv[1]: name of 8 bit uncompressed TGA to read
  argv[2]: name of C64 font image to write

  ASM listing of charset is also dumped to stdout.
  Use >.

  Wotan mit uns.
*/

#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>


#ifdef NDEBUG
#define IFNOTINITIALIZED(s)
#define IFINITIALIZED(s)
#else
#define IFNOTINITIALIZED(s) if (!initialized) SystemExit(ENOMEM, "ERROR: Not initialized.\n%s\n", s);
#define IFINITIALIZED(s) if (initialized) SystemExit(ENOMEM, "ERROR: Not initialized.\n%s\n", s);
#endif


#ifndef FALSE
#define FALSE               0
#endif

#ifndef TRUE
#define TRUE                1
#endif


typedef int BOOL;
typedef unsigned char BYTE;             // 8 bit unsigned
typedef unsigned short WORD;            // 16 bit unsigned
typedef unsigned long DWORD;            // 32 bit unsigned


typedef struct {
  BYTE b;
  BYTE g;
  BYTE r;
} PALETTE;


#define TGA_HEADER_SIZE                 18

typedef struct {
  BYTE idfieldlen;
  BYTE colormaptype;
  BYTE imagetype;
  WORD firstcolor;
  WORD colorsnum;
  BYTE bitspercolor;
  WORD xorig;
  WORD yorig;
  WORD width;
  WORD height;
  BYTE bitsperpixel;
  BYTE desc;
} TGA_HEADER;


// Writes a message to stdout and exits with exit code.
void SystemExit(int exitcode, const char *message, ...)
{
  va_list va;
  static char buf[256];

  va_start(va, message);
  vsprintf(buf, message, va);
  va_end(va);
  fprintf(stderr, buf);
  exit(exitcode);
}


class TGAImage {
  BOOL initialized;

  TGA_HEADER tga;

  int imagetype;
  int width;
  int height;
  int bpp;
  int cmaplen;

  PALETTE pal[256];
  BYTE *pixels;

// File interface.
  FILE *f;

  long tell(void);
  int seek(long n);
  int advance(long n);

  BYTE getByte(void);
  WORD getWord(void);
  DWORD getDWord(void);

public:
  TGAImage(void);
  TGAImage(const char *fname);
  ~TGAImage(void);

  void open(const char *fname);
  void close(void);

// Return width, height, bits per pixel and amount of total memory required.
  int getWidth(void);
  int getHeight(void);
  int getBPP(void);

  void readIndexed(BOOL flip=FALSE);
  void writeC64Font(const char *fname);
};


TGAImage::TGAImage(void)
{
  initialized=FALSE;
}


TGAImage::TGAImage(const char *fname)
{
  initialized=FALSE;
  open(fname);
}


TGAImage::~TGAImage(void)
{
  close();
}


// Seek to a position.
//
//  Input:     n:       offset from beginning of file
// Output: [ret]:       0 if fseek() was OK, nonzero otherwise
int TGAImage::seek(long n)
{
  return fseek(f, n, SEEK_SET);
}


// Advance by n bytes.
//
//  Input:     n:       number of bytes to advance
// Output: [ret]:       0 if fseek() was OK, nonzero otherwise
int TGAImage::advance(long n)
{
  return fseek(f, n, SEEK_CUR);
}


// Read an unsigned byte from the file.
// Exits program if read failed.
//
// Output: [ret]:       unsigned byte read from current position
BYTE TGAImage::getByte(void)
{
  BYTE t;
  int r;

  r=fread(&t, sizeof(t), 1, f);
  if (r!=1) SystemExit(99, "ERROR: Unable to read from file.");
  return t;
}


// Read an unsigned word (16 bits) from the file.
// Exits program if read failed.
//
// Output: [ret]:       unsigned word read from current position
WORD TGAImage::getWord(void)
{
  WORD t;
  int r;

  r=fread(&t, sizeof(t), 1, f);
  if (r!=1) SystemExit(99, "ERROR: Unable to read from file.");
  return t;
}


// Read an unsigned double word (32 bits) from the file.
// Exits program if read failed.
//
// Output: [ret]:       unsigned double word read from current position
DWORD TGAImage::getDWord(void)
{
  DWORD t;
  int r;

  r=fread(&t, sizeof(t), 1, f);
  if (r!=1) SystemExit(99, "ERROR: Unable to read from file.");
  return t;
}


void TGAImage::open(const char *fname)
{
// Close previous file if the caller forgot.
  if (initialized) close();
// Open file and read header.
  f=fopen(fname, "rb");
  if (f==NULL) {
    fprintf(stderr, "Unable to open input.");
    exit(EXIT_FAILURE);
  }
  tga.idfieldlen=getByte();
  tga.colormaptype=getByte();
  tga.imagetype=getByte();
  tga.firstcolor=getWord();
  tga.colorsnum=getWord();
  tga.bitspercolor=getByte();
  tga.xorig=getWord();
  tga.yorig=getWord();
  tga.width=getWord();
  tga.height=getWord();
  tga.bitsperpixel=getByte();
  tga.desc=getByte();
// Extract relevant information from the header.
  imagetype=tga.imagetype;
  width=tga.width;
  height=tga.height;
  switch (imagetype) {
// Paletted file.
    case 1:
      bpp=8;
      cmaplen=tga.colorsnum;
      break;
// Greyscale file.
    case 3:
      bpp=8;
      cmaplen=256;
      break;
// Unknown file.
    default:
      SystemExit(EINVAL, "ERROR: File %s is unsupported type.\nvoid TGAImage::open(const char *fname)\n", fname);
  }
  pixels=new BYTE[width*height];
  initialized=TRUE;
}


void TGAImage::close(void)
{
// Destructor for the caller instance of TGAImage is called automatically.
// If the caller called TGAImage::close(void) properly,
// the destructor calling it again creates an error.
// For this reason the following line is commented out.
//  IFNOTINITIALIZED("void TGAImage::close(void)");
  if (initialized) {
    fclose(f);
    delete pixels;
  }
  initialized=FALSE;
}


int TGAImage::getWidth(void)
{
  IFNOTINITIALIZED("int TGAImage::getWidth(void)");
  return width;
}


int TGAImage::getHeight(void)
{
  IFNOTINITIALIZED("int TGAImage::getHeight(void)");
  return height;
}


int TGAImage::getBPP(void)
{
  IFNOTINITIALIZED("int TGAImage::getBPP(void)");
  return bpp;
}


void TGAImage::readIndexed(BOOL flip)
{
  int bitmapsrc, palsrc;
  int i;
  int r, g, b, t, x, y, d;

  switch (imagetype) {
// Paletted file.
    case 1:
      palsrc=TGA_HEADER_SIZE+tga.idfieldlen;
      seek(palsrc);
      for (i=0; i<tga.colorsnum; i++) {
        switch (tga.bitspercolor) {
          case 16:
            t=getWord();
            r=((t>>10)&0x1f)<<3;
            g=((t>>5)&0x1f)<<3;
            b=(t&0x1f)<<3;
            break;
          case 24:
            b=getByte();
            g=getByte();
            r=getByte();
            break;
          case 32:
            b=getByte();
            g=getByte();
            r=getByte();
            getByte();
            break;
          default:              // This is impossible, just shut up compiler.
            b=getByte();
            g=getByte();
            r=getByte();
            getByte();
        }
        pal[i].r=(BYTE)r;
        pal[i].g=(BYTE)g;
        pal[i].b=(BYTE)b;
      }
      bitmapsrc=TGA_HEADER_SIZE+tga.idfieldlen+tga.colorsnum*tga.bitspercolor/8;
      seek(bitmapsrc);
      for (y=0; y<height; y++) {
// Is image flipped?
        if ((tga.desc&0x20==0x20)^flip) d=y;
        else d=height-1-y;
        for (x=0; x<width; x++) {
          t=getByte();
          pixels[d*width+x]=(BYTE)t;
        }
      }
      break;
// Grayscale file.
    case 3:
      bitmapsrc=TGA_HEADER_SIZE+tga.idfieldlen+tga.colorsnum*tga.bitspercolor/8;
      for (y=0; y<height; y++) {
// Is image flipped?
        if ((tga.desc&0x20==0x20)^flip) d=y;
        else d=height-1-y;
        for (x=0; x<width; x++) {
          t=getByte();
          pixels[d*width+x]=(BYTE)t;
        }
      }
      break;
  }
}


void TGAImage::writeC64Font(const char *fname)
{
  FILE *out;
  int xc, yc;                           // Character positions.
  int xp, yp;                           // Pixel positions.
  int m;                                // Bit mask for current x pixel pos.
  int t;                                // Bits collected so far.
  int i;

  out=fopen(fname, "wb");
  if (f==NULL) {
    fprintf(stderr, "Unable to open output.");
    exit(EXIT_FAILURE);
  }
  for (yc=0; yc<height; yc+=8) {
    for (xc=0; xc<width; xc+=8) {
      for (yp=0; yp<8; yp++) {
        m=128;
        t=0;
        for (xp=0; xp<8; xp++) {
          i=(yc+yp)*width+xc+xp;
          if (pixels[i]) t|=m;
          m>>=1;
        }
        fwrite(&t, 1, 1, out);
        if (yp==0) printf("        .byte ");
        printf("$%02x", t);
        if (yp!=7) printf(",");
        else printf("   ; $%04x\n", (width/8)*(yc/8)+xc/8);
      }
    }
  }
  fclose(out);
}


int main(int argc, char *argv[])
{
  TGAImage *im;

  if (argc==3) {
    im=new TGAImage();
    im->open(argv[1]);
    im->readIndexed();
    im->writeC64Font(argv[2]);
    delete im;
  }
  return EXIT_SUCCESS;
}
